==> SNPlocs.Hsapiens.dbSNP.20100427_0.99.6.tar.gz: SNPlocs.Hsapiens.dbSNP.20100427/R/getSNPlocs.R <==
### =========================================================================
### Loading the SNPs
### -------------------------------------------------------------------------

getSNPcount <- function()
{
    objname <- "SNPcount"
    ans <- getdata(objname)
    if (!is.integer(ans) || !identical(names(ans), SEQNAMES))
        stop("internal error: '", objname, "' data set is broken.\n",
             "       Please contact the maintainer of the ",
             PKGNAME, " package.")
    ans
}

### Load raw snplocs.
.load_snplocs <- function(seqname, caching)
{
    objname <- paste(seqname, "_snplocs", sep="")
    ans <- getdata(objname, caching=caching)
    empty_snplocs <- getdata("empty_snplocs")
    if (!identical(sapply(ans, class), sapply(empty_snplocs, class)))
        stop("internal error: unexpected col names and/or col types\n",
             "       for the '", objname, "' data set.\n",
             "       Please contact the maintainer of the ",
             PKGNAME, "\n       package.")
    if (nrow(ans) != getSNPcount()[seqname])
        stop("internal error: nb of rows in object '", objname, " ",
             "doesn't match the\n       nb of SNPs reported by ",
             PKGNAME, ":::getSNPlocs().\n",
             "       Please contact the maintainer of the ",
             PKGNAME, "\n       package.")
    ans
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### For SNP injection.
###

.loadLoc <- function(seqname) .load_snplocs(seqname, TRUE)$loc

.loadAlleles <- function(seqname) .load_snplocs(seqname, TRUE)$alleles


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### For the end-user.
###

.get_rsid_offsets <- function()
{
    offsets <- c(0L, cumsum(getSNPcount()))
    offsets <- offsets[-length(offsets)]
    names(offsets) <- names(getSNPcount())
    offsets
}

### Load rs ids for a given sequence. Return them in an integer vector.
.load_rsids <- function(seqname)
{
    all_rsids <- getdata("all_rsids")
    seq_pos <- match(seqname, names(getSNPcount()))
    offset <- .get_rsid_offsets()[seq_pos]
    idx <- seq_len(getSNPcount()[seq_pos]) + offset
    all_rsids[idx]
}

### Get user-friendly snplocs.
.get_ufsnplocs <- function(seqname, caching)
{
    rsids <- as.character(.load_rsids(seqname))
    snplocs <- .load_snplocs(seqname, caching)
    alleles <- safeExplode(rawToChar(snplocs$alleles))
    data.frame(RefSNP_id=rsids,
               alleles_as_ambig=alleles,
               loc=snplocs$loc,
               stringsAsFactors=FALSE)
}

.SNPlocsAsGranges <- function(ufsnplocs, seqname)
{
    if (is(seqname, "Rle")) {
        if (length(seqname) != nrow(ufsnplocs)
         || !identical(levels(seqname), SEQNAMES))
            stop("when an Rle, 'seqname' must be a factor Rle ",
                 "of length 'nrow(ufsnplocs)' and levels 'SEQNAMES'")
        ans_seqnames <- seqname
    } else {
        if (!is.factor(seqname))
            seqname <- factor(seqname, levels=SEQNAMES)
        if (length(seqname) == 1L)
            ans_seqnames <- Rle(seqname, nrow(ufsnplocs))
        else if (length(seqname) == nrow(ufsnplocs))
            ans_seqnames <- Rle(seqname)
        else 
            stop("'length(seqname)' must be 1 or 'nrow(ufsnplocs)'")
    }
    if (nrow(ufsnplocs) == 0L)
        ans_ranges <- IRanges()
    else
        ans_ranges <- IRanges(start=ufsnplocs$loc, width=1L)
    ans_strand <- Rle(strand("*"), nrow(ufsnplocs))
    GRanges(seqnames=ans_seqnames,
            ranges=ans_ranges,
            strand=ans_strand,
            RefSNP_id=ufsnplocs$RefSNP_id,
            alleles_as_ambig=ufsnplocs$alleles_as_ambig)
}

### Returns a data frame (as.GRanges=FALSE) or a GRanges object
### (as.GRanges=TRUE).
getSNPlocs <- function(seqname, as.GRanges=FALSE, caching=TRUE)
{
    if (!is.character(seqname) || any(is.na(seqname)))
        stop("'seqname' must be a character vector with no NAs")
    if (!all(seqname %in% SEQNAMES))
        stop("all 'seqname' elements must be in: ",
             paste(SEQNAMES, collapse=", "))
    if (!isTRUEorFALSE(as.GRanges))
        stop("'as.GRanges' must be TRUE or FALSE")
    if (!isTRUEorFALSE(caching))
        stop("'caching' must be TRUE or FALSE")
    if (!as.GRanges) {
        if (length(seqname) != 1L)
            stop("'seqname' must be of length 1 when 'as.GRanges' is FALSE")
        return(.get_ufsnplocs(seqname, caching))
    }
    if (length(seqname) == 0L) {
        empty_ufsnplocs <- getdata("empty_ufsnplocs")
        return(.SNPlocsAsGranges(empty_ufsnplocs, character(0)))
    }
    list_of_ufsnplocs <- lapply(seqname, .get_ufsnplocs, caching)
    ufsnplocs <- do.call(rbind, list_of_ufsnplocs)
    seqnames <- Rle(factor(seqname, levels=SEQNAMES),
                    unlist(lapply(list_of_ufsnplocs, nrow), use.names=FALSE))
    .SNPlocsAsGranges(ufsnplocs, seqnames)
}

.normargRsids <- function(rsids)
{
    if (!is.vector(rsids))
        stop("'rsids' must be an integer or character vector")
    if (IRanges:::anyMissing(rsids))
        stop("'rsids' cannot contain NAs")
    if (is.numeric(rsids)) {
        if (!is.integer(rsids))
            rsids <- as.integer(rsids)
        return(rsids)
    }
    if (!is.character(rsids))
        stop("'rsids' must be an integer or character vector")
    prefixes <- unique(substr(rsids, 1L, 2L))
    if ("rs" %in% prefixes) {
        if (!setequal(prefixes, "rs"))
            stop("'rsids' cannot mix ids that are prefixed with \"rs\" ",
                 "with ids that are not")
        ## Drop the "rs" prefix.
        rsids <- substr(rsids, 3, nchar(rsids))
    }
    rsids <- suppressWarnings(as.integer(rsids))
    if (IRanges:::anyMissing(rsids))
        stop("cannot extract the digital part of some ids in 'rsids'")
    rsids
}

### Returns a named integer vector where each (name, value) pair corresponds
### to a supplied rs id. The name is the chromosome of the rs id and the
### value is the row index in the serialized snplocs data frame corresponding
### to the rs id.
.rsid2rowidx <- function(rsids)
{
    if (length(rsids) == 0L) {
        idx <- integer(0)
    } else {
        all_rsids <- getdata("all_rsids")
        idx <- match(rsids, all_rsids)
        bad_rsids_idx <- which(is.na(idx))
        if (length(bad_rsids_idx) != 0L) {
            bad_rsids <- rsids[bad_rsids_idx]
            bad_rsids_in1string <- paste(bad_rsids, collapse=", ")
            stop("rs id(s) not found: ", bad_rsids_in1string)
        }
    }
    seqidx <- findInterval(idx - 1L, cumsum(getSNPcount())) + 1L
    rowidx <- idx - .get_rsid_offsets()[seqidx]
    names(rowidx) <- names(getSNPcount())[seqidx]
    rowidx
}

### Returns a named integer vector where each (name, value) pair corresponds
### to a supplied rs id. The name is the chromosome of the rs id and the
### value is its position on the chromosome.
rsid2loc <- function(rsids, caching=TRUE)
{
    rsids <- .normargRsids(rsids)
    if (!isTRUEorFALSE(caching))
        stop("'caching' must be TRUE or FALSE")
    rowidxs <- .rsid2rowidx(rsids)
    if (length(rowidxs) == 0L) {
        ans <- integer(0)
    } else {
        rowidxs_list <- split(unname(rowidxs), names(rowidxs))
        loc_list <- lapply(names(rowidxs_list),
                           function(seqname) {
                               rowidx <- rowidxs_list[[seqname]]
                               .load_snplocs(seqname, caching)$loc[rowidx]
                           })
        ans <- unsplit(loc_list, names(rowidxs))
    }
    names(ans) <- names(rowidxs)
    ans
}

### Returns a named character vector where each (name, value) pair corresponds
### to a supplied rs id. The name is the chromosome of the rs id and the
### value is a single IUPAC code representing the associated alleles.
rsid2alleles <- function(rsids, caching=TRUE)
{
    rsids <- .normargRsids(rsids)
    if (!isTRUEorFALSE(caching))
        stop("'caching' must be TRUE or FALSE")
    rowidxs <- .rsid2rowidx(rsids)
    if (length(rowidxs) == 0L) {
        ans <- raw(0)
    } else {
        rowidxs_list <- split(unname(rowidxs), names(rowidxs))
        alleles_list <- lapply(names(rowidxs_list),
                           function(seqname) {
                               rowidx <- rowidxs_list[[seqname]]
                               .load_snplocs(seqname, caching)$alleles[rowidx]
                           })
        ans <- unsplit(alleles_list, names(rowidxs))
    }
    ans <- safeExplode(rawToChar(ans))
    names(ans) <- names(rowidxs)
    ans
}

rsidsToGRanges <- function(rsids, caching=TRUE)
{
    rsids <- .normargRsids(rsids)
    if (!isTRUEorFALSE(caching))
        stop("'caching' must be TRUE or FALSE")
    loc <- rsid2loc(rsids, caching=caching)
    alleles <- rsid2alleles(rsids, caching=caching)
    ufsnplocs <- data.frame(RefSNP_id=as.character(rsids),
                            alleles_as_ambig=unname(alleles),
                            loc=unname(loc),
                            stringsAsFactors=FALSE)
    .SNPlocsAsGranges(ufsnplocs, names(loc))
}

==> SNPlocs.Hsapiens.dbSNP.20100427_0.99.7.tar.gz: SNPlocs.Hsapiens.dbSNP.20100427/R/getSNPlocs.R <==
### =========================================================================
### Loading the SNPs
### -------------------------------------------------------------------------

getSNPcount <- function()
{
    objname <- "SNPcount"
    ans <- getdata(objname)
    if (!is.integer(ans) || !identical(names(ans), SEQNAMES))
        stop("internal error: '", objname, "' data set is broken.\n",
             "       Please contact the maintainer of the ",
             PKGNAME, " package.")
    ans
}

### Load raw snplocs.
.load_snplocs <- function(seqname, caching)
{
    objname <- paste(seqname, "_snplocs", sep="")
    ans <- getdata(objname, caching=caching)
    empty_snplocs <- getdata("empty_snplocs")
    if (!identical(sapply(ans, class), sapply(empty_snplocs, class)))
        stop("internal error: unexpected col names and/or col types\n",
             "       for the '", objname, "' data set.\n",
             "       Please contact the maintainer of the ",
             PKGNAME, "\n       package.")
    if (nrow(ans) != getSNPcount()[seqname])
        stop("internal error: nb of rows in object '", objname, " ",
             "doesn't match the\n       nb of SNPs reported by ",
             PKGNAME, ":::getSNPlocs().\n",
             "       Please contact the maintainer of the ",
             PKGNAME, "\n       package.")
    ans
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### For SNP injection.
###

.loadLoc <- function(seqname) .load_snplocs(seqname, TRUE)$loc

.loadAlleles <- function(seqname) .load_snplocs(seqname, TRUE)$alleles


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### For the end-user.
###

.get_rsid_offsets <- function()
{
    offsets <- c(0L, cumsum(getSNPcount()))
    offsets <- offsets[-length(offsets)]
    names(offsets) <- names(getSNPcount())
    offsets
}

### Load rs ids for a given sequence. Return them in an integer vector.
.load_rsids <- function(seqname)
{
    all_rsids <- getdata("all_rsids")
    seq_pos <- match(seqname, names(getSNPcount()))
    offset <- .get_rsid_offsets()[seq_pos]
    idx <- seq_len(getSNPcount()[seq_pos]) + offset
    all_rsids[idx]
}

### Get user-friendly snplocs.
.get_ufsnplocs <- function(seqname, caching)
{
    rsids <- as.character(.load_rsids(seqname))
    snplocs <- .load_snplocs(seqname, caching)
    alleles <- safeExplode(rawToChar(snplocs$alleles))
    data.frame(RefSNP_id=rsids,
               alleles_as_ambig=alleles,
               loc=snplocs$loc,
               stringsAsFactors=FALSE)
}

.SNPlocsAsGranges <- function(ufsnplocs, seqname)
{
    if (is(seqname, "Rle")) {
        if (length(seqname) != nrow(ufsnplocs)
         || !identical(levels(seqname), SEQNAMES))
            stop("when an Rle, 'seqname' must be a factor Rle ",
                 "of length 'nrow(ufsnplocs)' and levels 'SEQNAMES'")
        ans_seqnames <- seqname
    } else {
        if (!is.factor(seqname))
            seqname <- factor(seqname, levels=SEQNAMES)
        if (length(seqname) == 1L)
            ans_seqnames <- Rle(seqname, nrow(ufsnplocs))
        else if (length(seqname) == nrow(ufsnplocs))
            ans_seqnames <- Rle(seqname)
        else 
            stop("'length(seqname)' must be 1 or 'nrow(ufsnplocs)'")
    }
    if (nrow(ufsnplocs) == 0L)
        ans_ranges <- IRanges()
    else
        ans_ranges <- IRanges(start=ufsnplocs$loc, width=1L)
    ans_strand <- Rle(strand("*"), nrow(ufsnplocs))
    GRanges(seqnames=ans_seqnames,
            ranges=ans_ranges,
            strand=ans_strand,
            RefSNP_id=ufsnplocs$RefSNP_id,
            alleles_as_ambig=ufsnplocs$alleles_as_ambig)
}

### Returns a data frame (as.GRanges=FALSE) or a GRanges object
### (as.GRanges=TRUE).
getSNPlocs <- function(seqname, as.GRanges=FALSE, caching=TRUE)
{
    if (!is.character(seqname) || any(is.na(seqname)))
        stop("'seqname' must be a character vector with no NAs")
    if (!all(seqname %in% SEQNAMES))
        stop("all 'seqname' elements must be in: ",
             paste(SEQNAMES, collapse=", "))
    if (!isTRUEorFALSE(as.GRanges))
        stop("'as.GRanges' must be TRUE or FALSE")
    if (!isTRUEorFALSE(caching))
        stop("'caching' must be TRUE or FALSE")
    if (!as.GRanges) {
        if (length(seqname) != 1L)
            stop("'seqname' must be of length 1 when 'as.GRanges' is FALSE")
        return(.get_ufsnplocs(seqname, caching))
    }
    if (length(seqname) == 0L) {
        empty_ufsnplocs <- getdata("empty_ufsnplocs")
        return(.SNPlocsAsGranges(empty_ufsnplocs, character(0)))
    }
    list_of_ufsnplocs <- lapply(seqname, .get_ufsnplocs, caching)
    ufsnplocs <- do.call(rbind, list_of_ufsnplocs)
    seqnames <- Rle(factor(seqname, levels=SEQNAMES),
                    unlist(lapply(list_of_ufsnplocs, nrow), use.names=FALSE))
    .SNPlocsAsGranges(ufsnplocs, seqnames)
}

.normargRsids <- function(rsids)
{
    if (!is.vector(rsids))
        stop("'rsids' must be an integer or character vector")
    if (S4Vectors:::anyMissing(rsids))
        stop("'rsids' cannot contain NAs")
    if (is.numeric(rsids)) {
        if (!is.integer(rsids))
            rsids <- as.integer(rsids)
        return(rsids)
    }
    if (!is.character(rsids))
        stop("'rsids' must be an integer or character vector")
    prefixes <- unique(substr(rsids, 1L, 2L))
    if ("rs" %in% prefixes) {
        if (!setequal(prefixes, "rs"))
            stop("'rsids' cannot mix ids that are prefixed with \"rs\" ",
                 "with ids that are not")
        ## Drop the "rs" prefix.
        rsids <- substr(rsids, 3, nchar(rsids))
    }
    rsids <- suppressWarnings(as.integer(rsids))
    if (S4Vectors:::anyMissing(rsids))
        stop("cannot extract the digital part of some ids in 'rsids'")
    rsids
}

### Returns a named integer vector where each (name, value) pair corresponds
### to a supplied rs id. The name is the chromosome of the rs id and the
### value is the row index in the serialized snplocs data frame corresponding
### to the rs id.
.rsid2rowidx <- function(rsids)
{
    if (length(rsids) == 0L) {
        idx <- integer(0)
    } else {
        all_rsids <- getdata("all_rsids")
        idx <- match(rsids, all_rsids)
        bad_rsids_idx <- which(is.na(idx))
        if (length(bad_rsids_idx) != 0L) {
            bad_rsids <- rsids[bad_rsids_idx]
            bad_rsids_in1string <- paste(bad_rsids, collapse=", ")
            stop("rs id(s) not found: ", bad_rsids_in1string)
        }
    }
    seqidx <- findInterval(idx - 1L, cumsum(getSNPcount())) + 1L
    rowidx <- idx - .get_rsid_offsets()[seqidx]
    names(rowidx) <- names(getSNPcount())[seqidx]
    rowidx
}

### Returns a named integer vector where each (name, value) pair corresponds
### to a supplied rs id. The name is the chromosome of the rs id and the
### value is its position on the chromosome.
rsid2loc <- function(rsids, caching=TRUE)
{
    rsids <- .normargRsids(rsids)
    if (!isTRUEorFALSE(caching))
        stop("'caching' must be TRUE or FALSE")
    rowidxs <- .rsid2rowidx(rsids)
    if (length(rowidxs) == 0L) {
        ans <- integer(0)
    } else {
        rowidxs_list <- split(unname(rowidxs), names(rowidxs))
        loc_list <- lapply(names(rowidxs_list),
                           function(seqname) {
                               rowidx <- rowidxs_list[[seqname]]
                               .load_snplocs(seqname, caching)$loc[rowidx]
                           })
        ans <- unsplit(loc_list, names(rowidxs))
    }
    names(ans) <- names(rowidxs)
    ans
}

### Returns a named character vector where each (name, value) pair corresponds
### to a supplied rs id. The name is the chromosome of the rs id and the
### value is a single IUPAC code representing the associated alleles.
rsid2alleles <- function(rsids, caching=TRUE)
{
    rsids <- .normargRsids(rsids)
    if (!isTRUEorFALSE(caching))
        stop("'caching' must be TRUE or FALSE")
    rowidxs <- .rsid2rowidx(rsids)
    if (length(rowidxs) == 0L) {
        ans <- raw(0)
    } else {
        rowidxs_list <- split(unname(rowidxs), names(rowidxs))
        alleles_list <- lapply(names(rowidxs_list),
                           function(seqname) {
                               rowidx <- rowidxs_list[[seqname]]
                               .load_snplocs(seqname, caching)$alleles[rowidx]
                           })
        ans <- unsplit(alleles_list, names(rowidxs))
    }
    ans <- safeExplode(rawToChar(ans))
    names(ans) <- names(rowidxs)
    ans
}

rsidsToGRanges <- function(rsids, caching=TRUE)
{
    rsids <- .normargRsids(rsids)
    if (!isTRUEorFALSE(caching))
        stop("'caching' must be TRUE or FALSE")
    loc <- rsid2loc(rsids, caching=caching)
    alleles <- rsid2alleles(rsids, caching=caching)
    ufsnplocs <- data.frame(RefSNP_id=as.character(rsids),
                            alleles_as_ambig=unname(alleles),
                            loc=unname(loc),
                            stringsAsFactors=FALSE)
    .SNPlocsAsGranges(ufsnplocs, names(loc))
}
